import java.util.ArrayList;
import java.util.Random;

import aiinterface.AIInterface;
import aiinterface.CommandCenter;
import enumerate.Action;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;

public class MytNoAI implements AIInterface {

	private Key key;
	private CommandCenter commandCenter;

	private GameData gameData;
	private boolean playerNumber;

	private FrameData frameData;

	private Simulator simulator;

	private ArrayList<Simulation> simulationsWithD_DF_FC;
	private ArrayList<Simulation> simulationsNoD_DF_FC;
	private ArrayList<Simulation> simulationsInSpeedRunningWithD_DF_FC;
	private ArrayList<Simulation> simulationsInSpeedRunningNoD_DF_FC;
	private int simulateIndex;
	private int simulateRepeat;
	private ArrayList<Simulation> simulationsUsing;

	private CharacterData self;
	private CharacterData opponent;

	private Random random;

	private Action[] actionToUse;
	private Action[] actionToUseInSpeedRunning;

	@Override
	public void close() {
		// TODO Auto-generated method stub

	}

	@Override
	public void getInformation(FrameData frameData) {
		// TODO Auto-generated method stub
		this.frameData = frameData;
		this.commandCenter.setFrameData(frameData, this.playerNumber);

		this.self = frameData.getCharacter(playerNumber);
		this.opponent = frameData.getCharacter(!playerNumber);
	}

	@Override
	public int initialize(GameData gameData, boolean playerNumber) {
		// TODO Auto-generated method stub
		this.gameData = gameData;
		this.playerNumber = playerNumber;

		this.actionToUse = new Action[] { Action.CROUCH_GUARD, Action.STAND, Action.FORWARD_WALK, Action.STAND_GUARD,
				Action.DASH, Action.JUMP, Action.STAND_A, Action.STAND_B, Action.STAND_FA, Action.STAND_FB,
				Action.THROW_A, Action.THROW_B, Action.CROUCH_A, Action.CROUCH_B, Action.CROUCH_FA, Action.CROUCH_FB,
				Action.STAND_D_DF_FC };

		this.actionToUseInSpeedRunning = new Action[] { Action.FORWARD_WALK, Action.STAND, Action.CROUCH_GUARD,
				Action.STAND_GUARD, Action.DASH, Action.JUMP, Action.STAND_A, Action.STAND_B, Action.STAND_FA,
				Action.STAND_FB, Action.THROW_A, Action.THROW_B, Action.CROUCH_A, Action.CROUCH_B, Action.CROUCH_FA,
				Action.CROUCH_FB, Action.STAND_D_DF_FC };

		this.simulationsWithD_DF_FC = new ArrayList<Simulation>();
		this.simulationsNoD_DF_FC = new ArrayList<Simulation>();
		this.simulationsInSpeedRunningWithD_DF_FC = new ArrayList<Simulation>();
		this.simulationsInSpeedRunningNoD_DF_FC = new ArrayList<Simulation>();

		for (Action action : this.actionToUse) {
			Simulation simulation = new Simulation(action);
			this.simulationsWithD_DF_FC.add(simulation);
		}

		for (int i = 0; i < this.actionToUse.length - 1; i++) {
			Simulation simulation = new Simulation(this.actionToUse[i]);
			this.simulationsNoD_DF_FC.add(simulation);
		}

		for (Action action : this.actionToUseInSpeedRunning) {
			Simulation simulation = new Simulation(action);
			this.simulationsInSpeedRunningWithD_DF_FC.add(simulation);
		}

		for (int i = 0; i < this.actionToUseInSpeedRunning.length - 1; i++) {
			Simulation simulation = new Simulation(this.actionToUseInSpeedRunning[i]);
			this.simulationsInSpeedRunningNoD_DF_FC.add(simulation);
		}

		this.key = new Key();
		this.commandCenter = new CommandCenter();

		this.simulator = this.gameData.getSimulator();

		this.random = new Random();

		gameData.getMotionData(playerNumber);
		gameData.getMotionData(!playerNumber);

		return 0;
	}

	@Override
	public Key input() {
		// TODO Auto-generated method stub
		return this.key;
	}

	@SuppressWarnings("deprecation")
	@Override
	public void processing() {
		// TODO Auto-generated method stub
		if (this.canProcessing()) {
			if (this.commandCenter.getSkillFlag()) {
				this.key = this.commandCenter.getSkillKey();
			} else {
				this.key.empty();
				this.commandCenter.skillCancel();

				this.simulateFrameData();

				this.resetScores();
				this.simulateIndex = 0;
				this.simulateRepeat = 0;

				if (this.self.getEnergy() >= 150 && this.gameData.getMaxHP(playerNumber) < 1000) {
					this.simulationsUsing = this.simulationsWithD_DF_FC;
				} else if (this.self.getEnergy() < 150 && this.gameData.getMaxHP(playerNumber) < 1000) {
					this.simulationsUsing = this.simulationsNoD_DF_FC;
				} else if (this.self.getEnergy() >= 150 && this.gameData.getMaxHP(playerNumber) >= 1000) {
					this.simulationsUsing = this.simulationsInSpeedRunningWithD_DF_FC;
				} else if (this.self.getEnergy() < 150 && this.gameData.getMaxHP(playerNumber) >= 1000) {
					this.simulationsUsing = this.simulationsInSpeedRunningNoD_DF_FC;
				}

				long startTime = System.nanoTime();
				while (System.nanoTime() - startTime < 16500000) {
					this.nextSimulation();
				}

				Action bestAction = this.getSimulationWithBestScore().action.getFirst();
				this.commandCenter.commandCall(bestAction.name());
			}
		}
	}

	@SuppressWarnings("deprecation")
	public boolean canProcessing() {
		return !this.frameData.getEmptyFlag() && this.frameData.getRemainingTime() > 0;
	}

	public void resetScores() {
		for (Simulation simulation : this.simulationsWithD_DF_FC) {
			simulation.score = 0.00;
		}
		for (Simulation simulation : this.simulationsNoD_DF_FC) {
			simulation.score = 0.00;
		}
	}

	public void nextSimulation() {
		Simulation simulation = new Simulation(Action.AIR_GUARD);

		boolean isRandoming = false;
		if (this.simulateIndex >= this.simulationsUsing.size() && this.simulateRepeat >= 3) {
			isRandoming = true;
			simulation = this.getSimulationWithBestScoreInSimulation();
		} else if (this.simulateIndex >= this.simulationsUsing.size() && this.simulateRepeat < 3) {
			this.simulateIndex = 0;
			this.simulateRepeat++;
			simulation = this.simulationsUsing.get(this.simulateIndex);
		} else {
			simulation = this.simulationsUsing.get(this.simulateIndex);
		}

		int time = 15;
		if (simulation.action.getFirst().name() == "STAND_D_DF_FC") {
			time = 40;
		}

		FrameData frameDataSimulated = this.simulator.simulate(this.frameData, this.playerNumber, simulation.action,
				null, time);

		int score;
		if (this.playerNumber) {
			score = (frameDataSimulated.getCharacter(true).getHp() - this.self.getHp())
					- (frameDataSimulated.getCharacter(false).getHp() - this.opponent.getHp());
		} else {
			score = (frameDataSimulated.getCharacter(false).getHp() - this.self.getHp())
					- (frameDataSimulated.getCharacter(true).getHp() - this.opponent.getHp());
		}

		simulation.score = (simulation.score * simulation.time + score) / (simulation.time + 1);
		simulation.time = simulation.time + 1;

		if (isRandoming == false) {
			this.simulateIndex++;
		} else {
			this.simulateIndex = 100;
		}
	}

	public Simulation getSimulationWithBestScore() {
		double bestScore = -10000;
		Simulation bestSimulation = new Simulation(Action.STAND_A);
		for (Simulation simulation : this.simulationsUsing) {
			if (simulation.score > bestScore) {
				bestScore = simulation.score;
				bestSimulation = simulation;
			}
			if (simulation.action.getFirst().name() == "STAND_D_DF_FC") {
				System.out.println(simulation.score);
			}
		}
		return bestSimulation;
	}

	public Simulation getSimulationWithBestScoreInSimulation() {
		double bestScore = -10000;
		int randomInt = 0;
		int bestIndex = 0;
		for (int index = 0; index < this.simulationsUsing.size(); index++) {
			Simulation simulation = this.simulationsUsing.get(index);
			if (bestScore < simulation.score) {
				bestScore = simulation.score;
				bestIndex = index;
			} else if (bestScore == simulation.score) {
				randomInt = this.random.nextInt(2);
				if (randomInt == 1) {
					bestScore = simulation.score;
					bestIndex = index;
				}
			}
		}
		Simulation bestSimulation = this.simulationsUsing.get(bestIndex);
		return bestSimulation;
	}

	public void simulateFrameData() {
		this.frameData = this.simulator.simulate(this.frameData, this.playerNumber, null, null, 15);
		if (playerNumber) {
			this.self = this.frameData.getCharacter(true);
			this.opponent = this.frameData.getCharacter(false);
		} else {
			this.self = this.frameData.getCharacter(false);
			this.opponent = this.frameData.getCharacter(true);
		}
	}

	@Override
	public void roundEnd(int arg0, int arg1, int arg2) {
		commandCenter.skillCancel();
		key.empty();
	}

}
